package xyz.matj1.bp;

import mcdr.sequence.impl.SequenceMocapPoseCoordsL2DTW;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import xyz.matj1.bp.enums.MocapType;
import xyz.matj1.bp.enums.SpeedMultiple;
import xyz.matj1.bp.experiments.KnnPrecision;
import xyz.matj1.bp.experiments.StretchMatching;

import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.regex.PatternSyntaxException;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static xyz.matj1.bp.enums.SpeedMultiple._1;

/**
 * A class for testing of how motion-word sequences
 * can be matched on artificially stretched sequences
 *
 * @author Matěj Bagar
 */
public class StretchedSequences {
	/**
	 * This runs tests on stretched sequences.
	 *
	 * @param args
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {

		// skeletons
		var baseSequenceMgmt = Utils.loadData(
			dataFileNames(MocapType.SKELETON).get(1d),
			MocapType.SKELETON.dataClass()
		);
		var stretchedSequenceMgmt = Utils.loadData(
			"data/skeleta/class130-actions-coords_normPOS-fps12_stretched-2×.data",
			SequenceMocapPoseCoordsL2DTW.class
		);
		var stretchMultiple = 2;
		var output = new File("output/graphs for stretched sequences/all_skeleta.pdf");

		// stretch matching
		if (false) {
			StretchMatching.show(
				StretchMatching.compute(baseSequenceMgmt, stretchedSequenceMgmt),
				stretchMultiple,
				output
			);
		}

		// kNN test
		if (false) {
			Utils.write(
				KnnPrecision.averageFormat(knnPrecisionAverage(KnnPrecision.compute(
					baseSequenceMgmt,
					stretchedSequenceMgmt
				))),
				new OutputStreamWriter(System.out)
			);
		}

		// kNN algorism correctness check
		{
			assert knnPrecisionAverage(KnnPrecision.compute(
				baseSequenceMgmt,
				baseSequenceMgmt
			)).equals(
				KnnPrecision.average(KnnPrecision.compute(
					baseSequenceMgmt
			)));
		}

		// TODO: udělat to přes různé délkové násobky
	}

	/**
	 * This returns a {@link Predicate} that tells if its given {@link LocalAbstractObject}
	 * has a locator URI matching to the locator URI of <code>sequence</code>.
	 * Two LocalAbstractObjects have matching locator URIs iff
	 * the URIs are digit sequences separated by “_” and the a three of them are equal.
	 * This is supposed to be used to search for matching stretched sequences.
	 *
	 * @param sequence a sequence to whose URI the other's URI shall be matched
	 * @return a predicate that tells if its given object has a locator URI matching to that of <code>sequence</code>
	 */
	public static Predicate<LocalAbstractObject> uriMatches(LocalAbstractObject sequence) {
		return other -> {
			try {
				var uriArray = sequence.getLocatorURI().split("_");
				var otherUriArray = other.getLocatorURI().split("_");

				return IntStream.of(0, 1, 2).allMatch(
					x -> uriArray[x].equals(otherUriArray[x])
				);
			} catch (PatternSyntaxException | NumberFormatException | IndexOutOfBoundsException e) {
				return false;
	}}; }

	/**
	 * This takes the output from {@link KnnPrecision#compute(ObjectMgmt, ObjectMgmt)}
	 * and creates an average of precision for each category.
	 *
	 * It's very similar to {@link KnnPrecision#average(Map)};
	 * the only difference is that that function takes data containing the sequences and ignores them,
	 * and this function takes data without the sequences.
	 *
	 * @param knnData an output from {@link KnnPrecision#compute(ObjectMgmt, ObjectMgmt)}
	 * @return a map from each category to the average precision in the category
	 */
	public static Map<Category, Double> knnPrecisionAverage(Map<Category, List<Double>> knnData) {
		return knnData.entrySet().parallelStream()
			.map(
				entry -> Map.entry(
					entry.getKey(),
					entry.getValue().parallelStream().mapToDouble(x -> x).average().orElse(Double.NaN)
			))
			.collect(Collectors.toConcurrentMap(Entry::getKey, Entry::getValue));
	}


	private static Map<SpeedMultiple, String> dataFileNames(MocapType type) {
		var speedMultiples = Arrays.asList(SpeedMultiple.values());

		return switch (type) {
			case SKELETON -> speedMultiples.stream().collect(Collectors.toUnmodifiableMap(
				Function.identity(),
				mult -> (
					"data/skeleta/class130-actions-coords_normPOS-fps12"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));

			case HARD_MW -> speedMultiples.stream().collect(Collectors.toUnmodifiableMap(
				Function.identity(),
				mult -> (
					"data/motion words/hard_motion_words"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));

			case SOFT_MW -> speedMultiples.stream().collect(Collectors.toUnmodifiableMap(
				Function.identity(),
				mult -> (
					"data/motion words/soft_motion_words"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));

			case MATCH_N_MW -> speedMultiples.stream().collect(Collectors.toUnmodifiableMap(
				Function.identity(),
				mult -> (
					"data/motion words/multi-overlay_motion_words"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));
	};}
}
