package xyz.matj1.bp;

import mcdr.objects.impl.ObjectMotionWord;
import mcdr.sequence.SequenceMotionWords;
import mcdr.test.utils.ObjectCategoryMgmt;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartTheme;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.util.ExportUtils;
import xyz.matj1.bp.exceptions.CacheIOException;
import xyz.matj1.bp.themes.XyGraphTheme;
import xyz.matj1.bp.types.Vector2;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.stream.Stream;

public class Utils {
	/**
	 * Loads data for testing motion words (and other sequences) from the file at the given path.
	 *
	 * @param path the given path
	 * @param class_ a class to which the sequences shall be castable
	 * @return an object with motion sequences and their categories
	 * @throws IOException if the file couldn't be read
	 */
	public static ObjectMgmt loadData(
		String path,
		Class<? extends LocalAbstractObject> class_
	) throws IOException {
		var mwCategoryMgmt = new ObjectCategoryMgmt();
		var sequenceMgmt = new ObjectMgmt(mwCategoryMgmt);
		sequenceMgmt.read(class_, path);

		return sequenceMgmt;
	}
	public static ObjectMgmt loadData(
		File path,
		Class<? extends LocalAbstractObject> class_
	) throws IOException {
		return loadData(path.getPath(), class_);
	}

	/**
	 * Returns a string representing the given ObjectMotionWord.
	 *
	 * @param seq the sequence castable to ObjectMotionWord
	 * @return a string representing the input sequence
	 */
	public static String sequenceToString(LocalAbstractObject seq) {
		StringBuilder rtv = new StringBuilder();
		for (ObjectMotionWord mw : ((SequenceMotionWords<? extends ObjectMotionWord>)seq).getSequenceData()) {
			if (rtv.length() > 0) {
				rtv.append(',');
			}
			rtv.append(mw.toString());
		}
		return seq.getLocatorURI() + " [" + rtv + "]";
	}

	/**
	 * Calculates the distance of two categories as the average
	 * of the distances of objects in each tuples from the Cartesian product of given collections.
	 *
	 * @param objectsInCategory1 a collection of the objects in the first category
	 * @param objectsInCategory2 a collection of the objects in the second category
	 * @return the distance of the categories
	 */
	public static Double distanceOfCategories(
		Collection<LocalAbstractObject> objectsInCategory1,
		Collection<LocalAbstractObject> objectsInCategory2
	) {
		ArrayList<Float> objectDistances = new ArrayList<>();
		for (LocalAbstractObject object1 : objectsInCategory1) {
			for (LocalAbstractObject object2 : objectsInCategory2) {
				objectDistances.add(object1.getDistance(object2));
			}
		}
		return averageOfFloats(objectDistances);
	}

	/**
	 * Returns the average of all the values in the input collection.
	 *
	 * @param collection the input collection
	 * @return average of the values in the input collection
	 */
	public static Double average(Collection<Double> collection) {
		return collection.parallelStream().mapToDouble(x -> x).average().orElse(Double.NaN);
	}

	/**
	 * Returns the average of all the values in the input collection.
	 *
	 * @param collection the input collection
	 * @return average of the values in the input collection
	 */
	public static Double averageOfFloats(Collection<Float> collection) {
		return collection.parallelStream().mapToDouble(x -> (double)x).average().orElse(Double.NaN);
	}

	/**
	 * Converts the given {@link Iterator} to {@link Stream} by putting each object returned from the iterator to the stream.
	 *
	 * @param iterator the given iterator
	 * @return a stream with the values taken from the iterator
	 */
	public static <T> Stream<T> iteratorToStream(Iterator<T> iterator) {
		Stream.Builder<T> builder = Stream.builder();
		iterator.forEachRemaining(builder::add);
		return builder.build().parallel();
	}

	/**
	 * Returns the length in poses of the given motion words sequence.
	 *
	 * @param object a motion words sequence
	 * @return the length of object
	 * @throws IllegalArgumentException if the given object isn't a motion words sequence
	 */
	public static int getLength(LocalAbstractObject object) throws IllegalArgumentException {
		try {
			return Integer.parseInt(object.getLocatorURI().split("_")[3]);
		} catch (IndexOutOfBoundsException | NumberFormatException e) {
			throw new IllegalArgumentException("The given object probably isn't a motion words sequence", e);
		}
	}

	/**
	 * This normalises a <code>value</code> in a way that,
	 * if its absolute value is smaller than 1, this returns 1/<code>value</code>,
	 * otherwise it returns <code>value</code>.
	 *
	 * @param value a value to be normalised
	 * @return <code>value</code> normalised
	 */
	public static double normaliseQuotient(double value) {
		return (Math.abs(value) < 1)
			? 1 / value
			: value;
	}

	/**
	 * This converts a Double to its string representation with <code>decimals</code> decimal places
	 * and ‘.’ as the separator of the whole part and the fractional part.
	 *
	 * @param input what shall be represented
	 * @param decimals decimal places of the representation
	 * @return a String representation of the given Double
	 */
	public static String displayDouble(Double input, int decimals) {
		double multiplier = Math.pow(10, decimals);
		return ((Double)(Math.round(input * multiplier) / multiplier)) // round to `decimals` decimal places
			.toString().replaceAll("\\.?0*$", ""); // strip maybe ‘.’ and ‘0’s from the end
	}

	/**
	 * This prints given string to given writers.
	 *
	 * @param data what shall be printed
	 * @param outputs where it shall be printed
	 * @throws IOException if writing to a given writer throws this exception
	 */
	public static void write(String data, Writer... outputs) throws IOException {
		for (var output : outputs) {
			output.write(data);
			output.flush();
		}
	}

	/**
	 * This creates a scatter-plot graph with the default theme
	 * ({@link XyGraphTheme}) and given format.
	 * The format of the output is PDF.
	 * This may change the given <code>chart</code> if the <code>chart</code> can't be easily cloned.
	 *
	 * @param chart a chart containing the data for the graph
	 * @param output where the graph shall be saved
	 * @param size the width and the height of the output (in pixels)
	 */
	public static void writeXYGraphAsPdf(
		JFreeChart chart,
		File output,
		Vector2<Integer> size
	) {
		writeXYGraphAsPdf(chart, output, size, XyGraphTheme.instance());
	}

	/**
	 * This creates a scatter-plot graph with a given theme and format.
	 * The format of the output is PDF.
	 * This may change the given <code>chart</code> if the <code>chart</code> can't be easily cloned.
	 *
	 * @param chart a chart containing the data for the graph
	 * @param output where the graph shall be saved
	 * @param size the width and the height of the output (in pixels)
	 */
	public static void writeXYGraphAsPdf(
		JFreeChart chart,
		File output,
		Vector2<Integer> size,
		ChartTheme theme
	) {
		try {
			chart = (JFreeChart)chart.clone();
		} catch (CloneNotSupportedException | ClassCastException ignored) {}

		theme.apply(chart);

		assert ExportUtils.isOrsonPDFAvailable();
		ExportUtils.writeAsPDF(chart, size.a(), size.b(), output);
	}

	/**
	 * This superimposes all datasets from the charts in <code>charts</code> to a single chart.
	 * So this adds all datasets from XYPlots of the charts to a single chart and returns the chart.
	 *
	 * The returned chart is a clone of the first chart from the iterator of <code>charts</code>.
	 * But, if it doesn't support cloning, the returned chart is the first chart from the iterator of <code>charts</code>.
	 * That means that the chart is changed
	 *
	 * @param charts a nonempty collection of charts
	 * @return a clone (or maybe original) of the first chart with added datasets of all other charts
	 * @throws IllegalArgumentException if <code>charts</code> is empty
	 */
	public static JFreeChart superimpose(Collection<JFreeChart> charts) {
		if (charts.isEmpty()) {
			throw new IllegalArgumentException("The collection of charts is empty.");
		}

		var chartIterator = charts.iterator();
		var firstChart = chartIterator.next();

		JFreeChart tempNewChart;
		try {
			tempNewChart = (JFreeChart)firstChart.clone();
		} catch (CloneNotSupportedException e) {
			tempNewChart = firstChart;
		}
		final var newChart = tempNewChart;

		var newPlot = newChart.getXYPlot();

		chartIterator.forEachRemaining(chart -> {
			var plot = chart.getXYPlot();
			for (int i = 0; i < plot.getDatasetCount(); i++) {
				newPlot.setDataset(newPlot.getDatasetCount(), plot.getDataset(i));
		}});

		return newChart;
	}

	/**
	 * This tries to read cache from the file on the given path and return them cast to <code>C</code>.
	 *
	 * @param cachePath a path to the file where to read from
	 * @param dataType the expected data type of read cache
	 * @param <C> the expected data type of read cache or its extension or implementation
	 * @return the data in the cache
	 * @throws CacheIOException if reading the file fails or if its data aren't serialisation of an object of type C
	 */
	public static <C> C readCache(String cachePath, Class<C> dataType) throws CacheIOException {
		return readCache(new File(cachePath), dataType);
	}

	/**
	 * This tries to read cache from the given file and return them cast to <code>C</code>.
	 *
	 * @param cacheFile where to read from
	 * @param dataType the expected data type of read cache
	 * @param <C> the expected data type of read cache or its extension or implementation
	 * @return the data in the cache
	 * @throws CacheIOException if reading the file fails or if its data aren't serialisation of an object of type C
	 */
	public static <C> C readCache(File cacheFile, Class<? extends C> dataType) throws CacheIOException {
		C data = null;
		try {
			var deserialiser = new ObjectInputStream(new FileInputStream(cacheFile));
			data = (C)deserialiser.readObject();
			deserialiser.close();
			System.err.printf("Cached data from %s were read successfully.%n", cacheFile);
			return data;
		} catch (IOException | ClassNotFoundException e) {
			System.err.printf(
				"There was an attempt to read cache from %s, but it wasn't read and an exception occurred:%n%s%n",
				cacheFile.getPath(), e
			);
			throw new CacheIOException(e);
	}}

	/**
	 * This writes serialises <code>data</code> and writes the result to the file at <code>cachePath</code>.
	 *
	 * @param cachePath a path to the file where to write the cache
	 * @param data the data to be cached
	 * @throws CacheIOException if the writing fails or the data can't be serialised
	 */
	public static void writeCache(String cachePath, Object data) throws CacheIOException {
		writeCache(new File(cachePath), data);
	}

	public static void writeCache(File cacheFile, Object data) throws CacheIOException {
		try {
			var serialiser = new ObjectOutputStream(new FileOutputStream(cacheFile));
			serialiser.writeObject(data);
			serialiser.close();
			System.err.printf("Data cache was written to %s.%n", cacheFile);
		} catch (IOException e) {
			System.err.printf(
				"There was an attempt to write cache to %s, but it wasn't written and an exception occurred:%n%s%n",
				cacheFile.getPath(), e
			);
			throw new CacheIOException(e);
		}
	}
}
