package xyz.matj1.bp.types;

import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.Function;

public abstract class Either<A, B> {
	public abstract Optional<A> left();
	public abstract Optional<B> right();

	public abstract <X> X either(Function<A, X> leftFunction, Function<B, X> rightFunction);
	public abstract <X, Y> Either<X, Y> bimap(Function<A, X> leftFunction, Function<B, Y> rightFunction);

	public boolean isLeft() {
		return this.left().isPresent();
	}
	public boolean isRight() {
		return this.right().isPresent();
	}

	public A leftOr(A fallback) {
		return left().orElse(fallback);
	}
	public B rightOr(B fallback) {
		return right().orElse(fallback);
	}

	public A leftOrThrow() throws NoSuchElementException {
		return this.left().orElseThrow(() -> new NoSuchElementException("This is a Left; it doesn't contain a right value."));
	}
	public B rightOrThrow() throws NoSuchElementException {
		return this.right().orElseThrow(() -> new NoSuchElementException("This is a Right; it doesn't contain a left value."));
	}

	@Override
	public final boolean equals(Object other) {
		if (other instanceof Either<?, ?> eOther) {
			if (this.isLeft()) {
				if (eOther.isLeft()){
					return this.leftOrThrow().equals(eOther.leftOrThrow());
			}} else {
				if (eOther.isRight()) {
					return this.rightOrThrow().equals(eOther.rightOrThrow());
		}}}

		return false;
	}

	@Override
	public int hashCode() {
		if (this.isLeft()) {
			return "left".hashCode() ^ this.left().hashCode();
		} else {
			return "right".hashCode() ^ this.right().hashCode();
	}}

	@Override
	public String toString() {
		var side = this.isLeft() ? "Left" : "Right";
		var content = this.isLeft() ? this.left().toString() : this.right().toString();
		return String.format("%s(%s)", side, content);
	}

	public static <X> X reduce(Either<X, X> e) {
		return e.either(Function.identity(), Function.identity());
	}

	public static class Left<A, B> extends Either<A, B> {
		private final A content;

		public Left(A content) {
			this.content = content;
		}

		@Override
		public Optional<A> left() {
			return Optional.of(content);
		}

		@Override
		public Optional<B> right() {
			return Optional.empty();
		}

		@Override
		public <X> X either(Function<A, X> leftFunction, Function<B, X> rightFunction) {
			return leftFunction.apply(this.content);
		}

		@Override
		public <X, Y> Either<X, Y> bimap(Function<A, X> leftFunction, Function<B, Y> rightFunction) {
			return new Left<X, Y>(leftFunction.apply(content));
		}
	}

	public static class Right<A, B> extends Either<A, B> {
		private final B content;

		public Right(B content) {
			this.content = content;
		}

		@Override
		public Optional<A> left() {
			return Optional.empty();
		}

		@Override
		public Optional<B> right() {
			return Optional.of(content);
		}

		@Override
		public <X> X either(Function<A, X> leftFunction, Function<B, X> rightFunction) {
			return rightFunction.apply(this.content);
		}

		@Override
		public <X, Y> Either<X, Y> bimap(Function<A, X> leftFunction, Function<B, Y> rightFunction) {
			return new Right<X, Y>(rightFunction.apply(content));
		}
	}
}
